-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local LibTSMTypes = select(2, ...).LibTSMTypes
local GroupExport = LibTSMTypes:DefineClassType("GroupExport")
local Table = LibTSMTypes:From("LibTSMUtil"):Include("Lua.Table")
local Operation = LibTSMTypes:Include("Operation")
local Group = LibTSMTypes:Include("Group")
local CustomString = LibTSMTypes:Include("CustomString")
local LibDeflate = LibStub("LibDeflate")
local LibSerialize = LibStub("LibSerialize")
local private = {
	isOperationSettingsTable = {},
	commonSettingKeys = {},
}
local MAGIC_STR = "TSM_EXPORT"
local VERSION = 1
local SERIALIZE_OPTIONS = {
	stable = true,
	filter = function(tbl, key)
		return not private.isOperationSettingsTable[tbl] or not private.commonSettingKeys[key]
	end,
}
local EMPTY_OPERATIONS = {}
local EMPTY_GROUP_OPERATIONS = {}
local EMPTY_CUSTOM_SOURCES = {}



-- ============================================================================
-- Static Class Functions
-- ============================================================================

---Creates a new group export object.
---@param ignoredOperationTypes string[] A list of operation types to exlucde from the export
---@param customStringSettings table<string, table<string, true|nil>> The operation settings which are custom strings
---@return GroupExport
function GroupExport.__static.New(ignoredOperationTypes, customStringSettings)
	if not next(private.commonSettingKeys) then
		for key in Operation.CommonSettingKeyIterator() do
			private.commonSettingKeys[key] = true
		end
	end
	local obj = GroupExport()
	obj:_Configure(ignoredOperationTypes, customStringSettings)
	return obj
end



-- ============================================================================
-- Class Meta Methods
-- ============================================================================

function GroupExport.__private:__init()
	self._ignoredOperationTypes = nil
	self._customStringSettings = nil
	self._groupPath = nil
	self._includedSubGroups = {}
	self._includeOperations = false
	self._includeCustomSources = false
	self._exportValid = false
	self._groups = {}
	self._items = {}
	self._operations = {}
	self._groupOperations = {}
	self._customSources = {}
	self._str = nil
end



-- ============================================================================
-- Public Class Methods
-- ============================================================================

---Resets the export context.
---@param groupPath GroupPathValue The group path to export
function GroupExport:Reset()
	assert(self._ignoredOperationTypes and self._customStringSettings)
	self._groupPath = nil
	wipe(self._includedSubGroups)
	self._includeOperations = false
	self._includeCustomSources = false
	self._exportValid = false
	for typeName in pairs(self._operations) do
		if not Operation.TypeExists(typeName) or self._ignoredOperationTypes[typeName] then
			self._operations[typeName] = nil
		end
	end
	for _, typeName in Operation.TypeIterator() do
		if not self._ignoredOperationTypes[typeName] then
			self._operations[typeName] = self._operations[typeName] or {}
			wipe(self._operations[typeName])
		end
	end
	wipe(self._groupOperations)
	wipe(self._customSources)
	self._str = nil
	collectgarbage()
end

---Sets the group path.
---@param groupPath GroupPathValue The group path to export
function GroupExport:SetPath(groupPath)
	assert(self._ignoredOperationTypes and self._customStringSettings)
	assert(Group.Exists(groupPath) and groupPath ~= Group.GetRootPath())
	if groupPath == self._groupPath then
		return
	end
	self._groupPath = groupPath
	self._exportValid = false
end

---Sets the subgroups to include in the export.
---@param subGroups table<GroupPathValue, true> Relative group paths to include as keys
function GroupExport:SetSubGroupsIncluded(subGroups)
	assert(self._ignoredOperationTypes and self._customStringSettings and self._groupPath)
	if Table.Equal(self._includedSubGroups, subGroups) then
		return
	end
	wipe(self._includedSubGroups)
	for relativeGroupPath in pairs(subGroups) do
		assert(Group.Exists(Group.JoinPath(self._groupPath, relativeGroupPath)))
		self._includedSubGroups[relativeGroupPath] = true
	end
	self._exportValid = false
end

---Sets whether or not operations are included in the export.
---@param includeOperations boolean Whether or not to include operations
---@param includeCustomSources boolean Whether or not to include custom sources (requires includeOperations)
function GroupExport:SetOperationsIncluded(includeOperations, includeCustomSources)
	assert(self._ignoredOperationTypes and self._customStringSettings and self._groupPath)
	assert(includeOperations or not includeCustomSources)
	assert(type(includeOperations) == "boolean" and type(includeCustomSources) == "boolean")
	if includeOperations ~= self._includeOperations then
		self._includeOperations = includeOperations
		self._str = nil
	end
	if includeCustomSources ~= self._includeCustomSources then
		self._includeCustomSources = includeCustomSources
		if not self._exportValid or next(self._customSources) then
			self._str = nil
		end
	end
end

---Gets the export string.
---@return string
function GroupExport:GetString()
	self:_Execute()
	return self._str
end

---Gets the number of items included in the export
---@return number
function GroupExport:GetNumItems()
	self:_Execute()
	return Table.Count(self._items)
end

---Gets the number of subgroups included in the export
---@return number
function GroupExport:GetNumSubGroups()
	self:_Execute()
	return Table.Count(self._groups) - 1
end

---Gets the number of operations included in the export
---@return number
function GroupExport:GetNumOperations()
	self:_Execute()
	local num = 0
	for _, moduleOperations in pairs(self._operations) do
		num = num + Table.Count(moduleOperations)
	end
	return num
end

---Gets the number of custom sources included in the export
---@return number
function GroupExport:GetNumCustomSources()
	self:_Execute()
	return Table.Count(self._customSources)
end



-- ============================================================================
-- Private Class Methods
-- ============================================================================

function GroupExport.__private:_Configure(ignoredOperationTypes, customStringSettings)
	assert(not self._ignoredOperationTypes and not self._customStringSettings)
	self._ignoredOperationTypes = {}
	for _, typeName in ipairs(ignoredOperationTypes) do
		self._ignoredOperationTypes[typeName] = true
	end
	self._customStringSettings = {}
	for typeName, settings in pairs(customStringSettings) do
		assert(not self._ignoredOperationTypes[typeName])
		self._customStringSettings[typeName] = {}
		for key, value in pairs(settings) do
			assert(type(key) == "string" and value == true)
			self._customStringSettings[typeName][key] = true
		end
	end
end

function GroupExport.__private:_Execute()
	assert(self._groupPath)
	if not self._exportValid then
		self:_Export()
		self._str = nil
	end
	if not self._str then
		self:_GenerateString()
	end
end

function GroupExport.__private:_Export()
	self:_ExportGroups()
	self:_ExportOperations()
	self:_ExportCustomSources()
	self._exportValid = true
end

function GroupExport.__private:_ExportGroups()
	wipe(self._groups)
	wipe(self._items)
	Group.Export(self._groupPath, self._includedSubGroups, self._groups, self._items)
end

function GroupExport.__private:_ExportOperations()
	for _, operations in pairs(self._operations) do
		wipe(operations)
	end
	wipe(self._groupOperations)
	for relGroupPath in pairs(self._groups) do
		local groupPath = relGroupPath == Group.GetRootPath() and self._groupPath or Group.JoinPath(self._groupPath, relGroupPath)
		self._groupOperations[relGroupPath] = {}
		for _, moduleName in Operation.TypeIterator() do
			if not self._ignoredOperationTypes[moduleName] then
				self._groupOperations[relGroupPath][moduleName] = {
					-- Always override at the top-level
					override = Group.HasOperationOverride(groupPath, moduleName) or groupPath == self._groupPath or nil,
				}
				for _, operationName in Group.OperationIterator(groupPath, moduleName) do
					Operation.UpdateFromRelationships(moduleName, operationName)
					if not Operation.IsIgnored(moduleName, operationName) then
						local operationSettings = Operation.GetSettings(moduleName, operationName)
						tinsert(self._groupOperations[relGroupPath][moduleName], operationName)
						self._operations[moduleName][operationName] = operationSettings
					end
				end
			end
		end
	end
	return self._operations, self._groupOperations
end

function GroupExport.__private:_ExportCustomSources()
	wipe(self._customSources)
	for moduleName, moduleOperations in pairs(self._operations) do
		for _, operationSettings in pairs(moduleOperations) do
			for key in pairs(self._customStringSettings[moduleName]) do
				self:_GetCustomSources(operationSettings[key])
			end
		end
	end
end

function GroupExport.__private:_GetCustomSources(str)
	for _, name, customSourceStr in CustomString.DependantCustomSourceIterator(str) do
		if not self._customSources[name] then
			self._customSources[name] = customSourceStr
			self:_GetCustomSources(customSourceStr)
		end
	end
end

function GroupExport.__private:_GenerateString()
	assert(not next(private.isOperationSettingsTable))
	for _, moduleOperations in pairs(self._operations) do
		for _, operationSettings in pairs(moduleOperations) do
			private.isOperationSettingsTable[operationSettings] = true
		end
	end
	local groupOperations = self._includeOperations and self._groupOperations or EMPTY_GROUP_OPERATIONS
	local operations = self._includeOperations and self._operations or EMPTY_OPERATIONS
	local customSources = self._includeCustomSources and self._customSources or EMPTY_CUSTOM_SOURCES
	local serialized = LibSerialize:SerializeEx(SERIALIZE_OPTIONS, MAGIC_STR, VERSION, Group.GetName(self._groupPath), self._items, self._groups, groupOperations, operations, customSources)
	wipe(private.isOperationSettingsTable)
	self._str = LibDeflate:EncodeForPrint(LibDeflate:CompressDeflate(serialized))
	collectgarbage()
end
