-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local LibTSMUtil = select(2, ...).LibTSMUtil
local EnumType = LibTSMUtil:Init("BaseType.EnumType")
local private = {
	typeStr = {},
	valueStr = {},
	valueTypeMap = {},
}
---@class EnumObject
---@class EnumValue



-- ============================================================================
-- Metatables
-- ============================================================================

local EQ_METHOD = function(obj1, obj2)
	-- Make sure they are both enum values
	if not private.valueStr[obj1] or not private.valueStr[obj2] then
		return false
	end
	-- Check if one is a parent of the other
	local type1 = private.valueTypeMap[obj1]
	local type2 = private.valueTypeMap[obj2]
	while type1 or type2 do
		if (type1 and rawequal(type1, obj2)) or (type2 and rawequal(type2, obj1)) then
			return true
		end
		type1 = type1 and private.valueTypeMap[type1]
		type2 = type2 and private.valueTypeMap[type2]
	end
	return false
end
local TYPE_MT = {
	__newindex = function()
		error("Enum is read-only")
	end,
	__index = function(_, key)
		error("Unknown enum value: "..tostring(key))
	end,
	__tostring = function(self)
		return private.typeStr[self]
	end,
	__eq = EQ_METHOD,
	__metatable = newproxy(false),
}
local VALUE_MT = {
	__newindex = function()
		error("Enum value is read-only")
	end,
	__index = function(_, key)
		error("Unknown enum value: "..tostring(key))
	end,
	__tostring = function(self)
		local str = private.valueStr[self]
		if not str then
			error("Invalid enum value")
		end
		return str
	end,
	__eq = EQ_METHOD,
	__metatable = newproxy(false)
}



-- ============================================================================
-- Module Functions
-- ============================================================================

---Creates an enum object.
---@generic T
---@param name string
---@param values T
---@return T
function EnumType.New(name, values)
	private.CreateType(name, values)
	for key, value in pairs(values) do
		private.TrackValue(value, key, values)
	end
	return values
end

---Creates a new enum value.
---@return EnumValue
function EnumType.NewValue()
	return setmetatable({}, VALUE_MT)
end

---Returns whether or not the passed value is an enum type.
---@param value EnumObject|any The value to check
---@return boolean
function EnumType.IsType(value)
	return private.typeStr[value] and not private.valueTypeMap[value] and type(value) == "table" and getmetatable(value) == TYPE_MT.__metatable
end

---Check if the value belongs to the specified enum type.
---@param value EnumValue|any The enum value
---@param enumType EnumObject The enum type
function EnumType.IsValue(value, enumType)
	assert(EnumType.IsType(enumType))
	-- Walk up the tree to handle nested enums
	local checkType = private.valueTypeMap[value]
	while checkType do
		if checkType == enumType then
			return true
		end
		checkType = private.valueTypeMap[checkType]
	end
	return false
end



-- ============================================================================
-- Private Helper Functions
-- ============================================================================

function private.CreateType(name, values)
	assert(type(name) == "string" and type(values) == "table")
	assert(not private.typeStr[values] and not private.valueTypeMap[values])
	private.typeStr[values] = name
	setmetatable(values, TYPE_MT)
end

function private.TrackValue(value, key, parentType)
	if type(value) == "table" and next(value) then
		-- Recurse for a nested type
		local name = tostring(parentType).."."..key
		private.CreateType(name, value)
		private.valueStr[value] = name
		private.valueTypeMap[value] = parentType
		for key2, value2 in pairs(value) do
			private.TrackValue(value2, key2, value)
		end
	else
		assert(type(key) == "string" and type(value) == "table")
		assert(getmetatable(value) == VALUE_MT.__metatable)
		assert(not private.valueStr[value] and not private.valueTypeMap[value])
		private.valueStr[value] = tostring(parentType).."."..key
		private.valueTypeMap[value] = parentType
	end
end
